// ************************************************************************************************
// Projet Vertical 2024 : maintenir vertical un pendule inversé monté sur un chariot horizontal
// ************************************************************************************************
// Fonctionne avec 2 autres Arduino, des nanos, chargés de compter les impulsions de 2 encodeurs rotatifs (chariot et barre)
//
// Type de carte : Arduino Uno
// COM7
// Régulation : LQR
//

// Wire.h : pour les liaison I2C
// Utilisée pour communiquer avec l'arduino nano "Angle", l'arduino nano "position chariot", et la carte de puissance PWM du moteur
#include <Wire.h>

#define ADR_ARDUINO_ANGLE 9      // Adresse I2C de l'arduino nano "Angle"
#define ADR_ARDUINO_XCHARIOT 8   // Adresse I2C de l'arduino nano "XChariot"

#define MODE_LQR "L"
#define MODE_MANUEL "M"

float angleMesure;   // angle mesuré en radian. 0=vertical
float angleMesurePrecedent = 0;
float vAngleMesure;  // Dérivée
float xChariotMesure;   // Position du chariot en m
float xChariotMesurePrecedent;
float vChariotMesure;  // Dérivée

String modeFonctionnement = MODE_MANUEL;

double k0 = -24.5, k1 = -22.0, k2 = 131, k3 = 9.1; //  <- adaptation manuelle des valeurs calculées : -22.4, -23.3, 102.8, 22.7
double u;            // commande du moteur (force)
double kU = 17.6;    // coeficient pour adapter u
int pMoteur = 0;     // 0=moteur arrêté

unsigned long currTime, prevTime = 0;
float dt;   // durée mesurée de la boucle principale

void setup() {
  Serial.begin(38400);
  Wire.begin();  // connexion au bus I2c en tant que maître
  delay(100);
  initMD22();
  delay(100);
  vitesseMoteur(pMoteur);
  envoiEtatSurSerial();
  // initialize_matrices();
}

void loop() {
  // Mesures des positions
  angleMesure = lectureAngle();
  xChariotMesure = lectureXChariot();
  // Mesure du temps dt
  currTime = millis();
  dt = currTime - prevTime; // Durée réelle de la boucle en ms. Environ 50ms.
  prevTime = currTime;
  // Calcul des vitesses
  vChariotMesure = (xChariotMesure - xChariotMesurePrecedent) / (dt / 1000);
  vAngleMesure = (angleMesure - angleMesurePrecedent) / (dt / 1000);
  xChariotMesurePrecedent = xChariotMesure;
  angleMesurePrecedent = angleMesure;

  if (modeFonctionnement == MODE_LQR) {
    u = - (k0 * xChariotMesure + k1 * vChariotMesure + k2 * angleMesure + k3 * vAngleMesure);  //  F(t) = -K.X(t)
    pMoteur = round(u * kU);
    vitesseMoteur(pMoteur); // entre -128 et +127
  }

  if (modeFonctionnement == MODE_MANUEL) {
    vitesseMoteur(pMoteur);
  }
  recoitPortSerie();
  envoiEtatSurSerial();
}

int angleAZero() {
  Wire.beginTransmission(ADR_ARDUINO_ANGLE);
  Wire.write("0");
  Wire.endTransmission();
}

// Angle de la barre, mesuré en radian. 0=vertical.
float lectureAngle() {
  byte angleByte0, angleByte1;
  int angle; // angle renvoyé par l'encodeur rotatif et son arduino nano "Angle" : 2400=1 tour=2.PI rad
  Wire.requestFrom(ADR_ARDUINO_ANGLE, 2);
  angleByte0 = Wire.read();
  angleByte1 = Wire.read();
  angle = angleByte0;
  angle = (angle << 8) | angleByte1;
  return -(angle / 2400.0) * 2 * PI;
}

// Position du chariot en mètre. 0=au centre
// Calibrage ; 5871 -> 0,506m
float lectureXChariot() {
  byte xChariotByte0, xChariotByte1;
  int xChariot; // position renvoyée par l'encodeur rotatif du moteur et son arduino nano "position chariot". Calibrage : 5871 -> 0,506m
  Wire.requestFrom(ADR_ARDUINO_XCHARIOT, 2);
  xChariotByte0 = Wire.read();
  xChariotByte1 = Wire.read();
  xChariot = xChariotByte0;
  xChariot = (xChariot << 8) | xChariotByte1;
  return (-xChariot / 5871.0 * 0.506);
}

void envoiEtatSurSerial() {
  Serial.print("Angle : " + String(angleMesure, 5) + "rad#");
  Serial.print("xChariot : " + String(xChariotMesure, 5) + "m#");
  Serial.print("kU : " + String(kU) + "#");
  Serial.print("pMoteur : " + String(pMoteur) + "#");
  Serial.print("dt : " + String(dt) + "ms#");
  Serial.print("(k0, k1, k2 k3)=(" + String(k0, 2) + ", " + String(k1, 2) + ", " + String(k2, 2) + ", " + String(k3, 2) + ")#");
  Serial.print("Mode Manuel/LQR : " + String(modeFonctionnement) + "#");
  Serial.println(""); // Fin de transmission
}
